C*****************************************
       program Kernel Density Approximation Graph 2D
C*****************************************
       implicit none
	   include 'common.fi'

       write(6,*) '================================================='
	   write(6,*) ' PROGRAM IS USED TO PLOT 2D GRAIN BOUNDARY PROPERTY '
	   write(6,*) ' DISTRIBUTIONS FROM LISTS OF GRAIN BOUNDARY '
	   write(6,*) ' TRIANGLES USING THE KERNEL DENISTY APPROXIMATION '
	   write(6,*) ' AS DESCRIBED BY GLOWINSKI AND MORAWEIC, 2014 '
	   write(6,*) ' http://dx.doi.org/10.1007/s11661-014-2325-y'
	   write(6,*) ' '
	   write(6,*) ' rohrer@cmu.edu'
	   write(6,*) ' version 06/19/2019'
       write(6,*) '================================================='
	   
	   version = 'version 06/19/2019'
       !This version reproduces results from D3D and the discrete binning
       !Method.  Thanks to Xiaoting Zhong for fixing the normalization
       !procedure and some of the other problems with the algorithm.

       !some constants that might be useful
       pi = 4.0*atan(1.0)
       !some formating statements used to format typed output
  5    format(A,A)
  6    format(A)
  7    format(A,I4,A,I4)
  14   format(A,I2,2X,A,I2)
  15   format(A,I7,A)
  16   format(A,I8,A,I9,A)
  20   format(A,A,A,A,f5.2,1x,f8.2,1x,f7.2,A)
  21   format(A,3I3,A,f5.2)
  22   format(A,A,A,A,f5.2,1x,f8.2,1x,f7.2)
  24   format(A,4(1x,f4.1))
  25   format(A,4(1x,f5.3))
  26   format(A,A,A,f4.2,1x,f4.2,1x,f4.2,A)

       open(21, file='input.txt', status='old') ! Open the file with the program
                                                ! parameters.
	   read(21,*) keyword                       ! This reads the filename with the data)
	   read(21,*) tempword, tempword, tempword  ! Reads a line containing headers
       read(21,*) msym, radian, dist_type       ! Specifies the symmetry, units for angles, plot type
       read(21,*) tempword                      ! Read a header line
	   read(21,*) rho_p                         ! Read the dimensions of the aperture
       read(21,*) name                          ! added to the output filename
       read(21,*) errors                        ! determines whether or not error estimates are written
       read(21,*) numDistinctGBs                ! number of distinct boundaries
       close(21)                                ! Close the input file

       point = index(keyword,'.') - 1

       !this updates the user about the parameters on the consol.
       write(6,5)'I am going to look for data in the file labeled: ', keyword
       write(6,14)'msym =',msym,'radian =',radian
       if (dist_type.eq.1) then
        write(6,6)'the input is being interpreted as grain boundary plane data'
       endif
       if (dist_type.eq.2) then
        write(6,6)'the input is being interpreted as grain boundary energy data'
       endif
       if (dist_type.eq.3) then
        write(6,6)'the input is being interpreted as grain boundary curvature data'
       endif

       ! This part determines the total number of lines
       ! in the data file and assigns it to the variable, nnline
		nnline = 0 ! Initialize counter
		open(41, file=keyword, status='old') ! open the file
 60     continue
		read(41,*,end=61) ! read line, if end of file, goto 61
		nnline=nnline+1  ! Increment counter
		goto 60
 61     continue
        close(41)  ! we are finished, close the file, report result.
		write(6,15) 'There are  ',nnline,' lines in the data file.'
		
        ! Check each line for the string '#'. If found, we know this line
        ! is part of the header.  If not, we have reached the data section.
	    open(41, file=keyword, status='old') ! open the file
        header = 0 ! Initialize counter
        hash = 0   ! Initialize flag
	    do i=1,nnline  ! loop over each line in the file
	     read(41, "(a)") inline    ! reads line into a string
	     hash = index(inline, '#') ! if the character "#" in the string, hash>1.
	     if (hash.ne.0) then       ! check the flag
	      header=header+1          ! when "#" found, it is a header line
		  hash=0                   ! reset the flag
	     else
		  goto 80                  ! when hash is = 0, we have reached the end of the header
	     endif
	    enddo  ! closes i=1,nnline
 80    close(41)				   ! close the file, report the results
	   write(6,15)'There are ',header,' lines in the header'
       write(6,15) 'There are  ',nnline-header,' triangles in the data file.'

       !This opens the output file, where results will be written
       open(30, file=keyword(1:point)//name//'_gmt1.gpf',status='unknown')
       !write(30,"(4(1x,f4.1))")  ax(1), ax(2), ax(3), ang
       !write(6,21) 'misorientation = [',int(ax(1)), int(ax(2)), int(ax(3)),'] ', ang
       !if data on errors was requested, a file is opened for this distribution
       if (errors.eq.1) then
        open(31, file=keyword(1:point)//name//'_errors_gmt1.gpf',status='unknown')
        !write(31,"(4(1x,f4.1))")  ax(1), ax(2), ax(3), ang
       else
        continue
       endif

       call get_symop (msym, O, nsymm)
       !call get_symop (msym, O, nsymm) calls the subroutine that contains the symmetry operators.
       !The variable msym, read from the input file, specifies the symmetry (1 = tetragonal,
       !2 = hexagonal, 3 = cubic, 4 = trigonal, 5 = orthorhombic), O is a matrix containing the
       !operators, and nsymm is the number of symmetry operators.  The routine returns O and nsymm

       !This is the factor needed for the normalization
       BallVolume = 2.0*nsymm*(1.0 - cos(rho_p*(pi/180.0)))
       rho_p=rho_p*(pi/180.0)  !We have to convert the aperture values to radians

       !read in the sampling points
       open(22, file='grid_points.txt',status='old')
       read(22,*)num_samplPts
       do i=1,num_samplPts
        read(22,*)samplPts(1,i),samplPts(2,i),samplPts(3,i)
       enddo
       close(22)

       do i=1,num_samplPts
        dist(i)=0.0
        err(i)=0.0
        num(i)=0.0
       enddo

       open(41, file=keyword, status='old') ! open the file and read the header
       do i=1,header
        read(41, "(a)") inline ! read a line of the header
       enddo! ends i=1,header loop

       write(6,6)'Reading data'
       totalArea = 0
       do i=header+1,nnline !read all data to the array tris
        if (dist_type.eq.1.or.dist_type.eq.2) then
         read(41,*)tris(i,1),tris(i,2),tris(i,3),tris(i,4),tris(i,5),tris(i,6),tris(i,7),tris(i,8),tris(i,9),tris(i,10)
        else
         read(41,*)tris(i,1),tris(i,2),tris(i,3),tris(i,4),tris(i,5),tris(i,6),tris(i,7),tris(i,8),tris(i,9),tris(i,10),tris(i,11)
        endif
        ! outputs from some programs produce zero Euler angles when
        ! when one or both orientiations are not known.  These lines are
        ! ignored because they will cause problems later.
        if (tris(i,1).eq.0.00.and.tris(i,2).eq.0.00.and.tris(i,3).eq.0.00) then
         goto 500 ! Sends it to the next line
        else
         continue
        endif
        if (tris(i,4).eq.0.00.and.tris(i,5).eq.0.00.and.tris(i,6).eq.0.00) then
         goto 500 ! Sends it to the next line
        else
         continue
        endif
        totalArea=totalArea+tris(i,10)
        !If the Euler angles are provided in units of degrees, these
        !lines convert them to radians
        if (radian.eq.0) then
         tris(i,1)=tris(i,1)*(pi/180.0)
         tris(i,2)=tris(i,2)*(pi/180.0)
         tris(i,3)=tris(i,3)*(pi/180.0)
         tris(i,4)=tris(i,4)*(pi/180.0)
         tris(i,5)=tris(i,5)*(pi/180.0)
         tris(i,6)=tris(i,6)*(pi/180.0)
        else
         continue
        endif
        if (dist_type.eq.2) then !This case is for a GB energy distribution
         tris(i,7)=tris(i,7)*(pi/180.0)
         tris(i,8)=tris(i,8)*(pi/180.0)
         call AnglesToV (tris(i,7), tris(i,8), normal_lab)
         tris(i,7)=normal_lab(1)
         tris(i,8)=normal_lab(2)
         tris(i,9)=normal_lab(3)
        endif


 500    continue
       enddo
       close(41)

       write(6,6)'Computing distribution'
       do j=1,num_samplPts  ! this is a loop over all grid points
        sum = 0.0 !zero the counters for each grid point
        c = 0.0
        numTris = 0.0
        do i=header+1,nnline
         e1(1)=tris(i,1)    ! here we assign the variables to their conventional names
         e1(2)=tris(i,2)    ! to be consistent with borrowed code below.
         e1(3)=tris(i,3)
         e2(1)=tris(i,4)
         e2(2)=tris(i,5)
         e2(3)=tris(i,6)
         normal_lab(1)=tris(i,7)
         normal_lab(2)=tris(i,8)
         normal_lab(3)=tris(i,9)
         if (dist_type.eq.1.or.dist_type.eq.2) then
          area = tris(i,10)
         else
          area = tris(i,11)
         endif

         !The Euler angles are converted to 3x3 orientation matrices, g_o1 and g_o2.
         call EToG(e1,g_o1)
         call EToG(e2,g_o2)
         fixedNormal1(1)=samplPts(1,j)
         fixedNormal1(2)=samplPts(2,j)
         fixedNormal1(3)=samplPts(3,j)


         do i_sy=1,nsymm             ! loop over all symmetry elements
          call MToV(g_o1, normal_lab, normal_grain1)
          call MToV(g_o2, normal_lab, normal_grain2)
          call symop(O,i_sy,so1)     ! get first symmetry element
          call MToV(so1, normal_grain1, sym_normal1)
          call MToV(so1, normal_grain2, sym_normal2)
          sign=1.0
          chi1=acos(sign*( (sym_normal1(1)*fixedNormal1(1)) + (sym_normal1(2)*fixedNormal1(2)) + (sym_normal1(3)*fixedNormal1(3)) ))
          chi2=acos(sign*( (sym_normal2(1)*fixedNormal1(1)) + (sym_normal2(2)*fixedNormal1(2)) + (sym_normal2(3)*fixedNormal1(3)) ))
          if (chi1.lt.rho_p) then
           numTris=numTris+1.0
           y = area - c
           t = sum + y
           c = (t - sum) - y
           sum = t
          endif
          if (chi2.lt.rho_p) then
           numTris=numTris+1.0
           sum = sum + area
          endif

          sign=-1.0
          chi1=acos(sign*( (sym_normal1(1)*fixedNormal1(1)) + (sym_normal1(2)*fixedNormal1(2)) + (sym_normal1(3)*fixedNormal1(3)) ))
          chi2=acos(sign*( (sym_normal2(1)*fixedNormal1(1)) + (sym_normal2(2)*fixedNormal1(2)) + (sym_normal2(3)*fixedNormal1(3)) ))
          if (chi1.lt.rho_p) then
           numTris=numTris+1.0
           y = area - c
           t = sum + y
           c = (t - sum) - y
           sum = t
          endif
          if (chi2.lt.rho_p) then
           numTris=numTris+1.0
           sum = sum + area
          endif

         enddo  !This closes the i_sy=1,nsymm loop

        enddo    ! This closes i=header+1,nnline loop

        dist(j)=sum ! assigns the total area at this sample point
        num(j)=numTris ! assigns number of energy values added at this sample point

        if (mod(j,20).eq.0) then
         write(6,7) 'At grid point ',j,' of ',num_samplPts
        endif

       enddo !j loop over sample points

       dist_max=0.0
       dist_min=10000000.0

       !normalize the distributions, and find the extreme values
       if (dist_type.eq.1) then !This part is for grain boundary plane distributions
        do i=1, num_samplPts
         dist(i)=dist(i)/totalArea
         dist(i) = dist(i)/BallVolume
         if (dist(i).gt.dist_max) then
          dist_max=dist(i)
         endif
         if (dist(i).lt.dist_min) then
          dist_min=dist(i)
         endif
        enddo
       else   !This part is for grain boundary energy/curvature distributions
        do i=1, num_samplPts
         dist(i)=dist(i)/num(i)
         if (dist(i).gt.dist_max) then
          dist_max=dist(i)
         endif
         if (dist(i).lt.dist_min) then
          dist_min=dist(i)
         endif
        enddo
       endif


         do i=1,num_samplPts
          vec(1)=samplPts(1,i)
          vec(2)=samplPts(2,i)
          vec(3)=samplPts(3,i)
          call VToAngles(vec,angles)
          theta_plot=angles(1)
          phi_plot=angles(2)
          write (30,*) phi_plot*(180.0/3.14159), (90.0-(theta_plot*(180.0/3.14159))), dist(i)
         enddo
         close(30) ! closes the output file

         if (errors.eq.1) then  ! Write a file with errors, when requested
          do i=1, num_samplPts
           err(i)=sqrt((dist(i)/totalArea/numDistinctGBs))/BallVolume
          enddo

          do i=1,num_samplPts
           vec(1)=samplPts(1,i)
           vec(2)=samplPts(2,i)
           vec(3)=samplPts(3,i)
           call VToAngles(vec,angles)
           phi_plot=angles(1)
           theta_plot=angles(2)
           write (31,*) phi_plot*(180.0/3.14159), (90.0-(theta_plot*(180.0/3.14159))), err(i)
          enddo
          close(31) ! closes the output file

         else
          continue
         endif

         interval = (dist_max-dist_min)/10.0 !Provide some guidance for setting the limits on the plot

         !this part outputs instructions for initiating Draw_stereograms.  One case for
         !hexagonal and another for cubic.
         if (msym.eq.2) then
          write(6,*)' '
          write(6,6)'to initiate the gmt script, enter ./Draw_stereograms [Number of plots] [filename]_gmt_ IF rainbow [min] [max] [interval] stereo HEX'
          write(6,20)'For example: ./Draw_stereograms 1 ',keyword(1:point),name,'_gmt IF rainbow ',dist_min,dist_max,interval,' stereo HEX'
         else
          write(6,*)' '
          write(6,6)'to initiate the gmt script, enter ./Draw_stereograms [Number of plots] [filename]_gmt_ IF rainbow [min] [max] [interval]'
          write(6,22)'For example: ./Draw_stereograms 1 ',keyword(1:point),name,'_gmt IF rainbow ',dist_min,dist_max,interval
         endif



 9000    continue

         end






	   
	   
	   
	   
	   
	   
	   
	   
	   
	   
	   
	   
	   
	   
	   
	   
	   
	   
	   
	   
	   