C*****************************************
       program Kernel Density Approximation Points
C*****************************************
       implicit none
	   include 'common.fi'

       write(6,*) '================================================='
	   write(6,*) ' PROGRAM IS USED TO COMPUTE GRAIN BOUNDARY  '
	   write(6,*) ' PROPERTIES AT A LIST OF USER SPECIFIED POINTS IN  '
	   write(6,*) ' THE 5D GB SPACE FROM FROM A SET OF GRAIN BOUNDARY  '
       write(6,*) ' TRIANGLES, USING THE KERNEL DENISTY APPROXIMATION'
	   write(6,*) ' AS DESCRIBED BY GLOWINSKI AND MORAWEIC, 2014 '
	   write(6,*) ' http://dx.doi.org/10.1007/s11661-014-2325-y'
	   write(6,*) ' '
	   write(6,*) ' rohrer@cmu.edu'
	   write(6,*) ' version 02/15/2021'
       write(6,*) '================================================='
	   
	   version = 'version 02/15/2021'
       ! version 05/24/2019: added curvature distributions
       ! version 06/10/2019: fixed small error in area summation and
       ! version 06/10/2019: added normalization for symmetry
       ! version 06/10/2019: added status update
       ! version 06/11/2019: read grid points from a list
       ! version 06/12/2019: converted from kernel density graph to read from list
       ! version 06/12/2019: finalized output
       ! version 02/15/2021: this was modified to deal with velocity
       ! version 02/15/2021: fixed acos to never return nan

       !some constants that might be useful
       pi = 4.0*atan(1.0)
       !some formating statements used to format typed output
  5    format(A,A)
  6    format(A)
  7    format(A,I4,A,I4)
  14   format(A,I2,2X,A,I2)
  15   format(A,I12,A)
  16   format(A,I8,A,I9,A)
  20   format(A,A,A,A,f5.2,1x,f8.2,1x,f7.2,A)
  21   format(A,3I3,A,f5.2)
  22   format(A,A,A,A,f5.2,1x,f8.2,1x,f7.2)
  24   format(A,4(1x,f4.1))
  25   format(A,4(1x,f7.3))
  26   format(A,A,A,f4.2,1x,f4.2,1x,f4.2,A)
  27   format(A,3(1x,f5.3))
  28   format(f6.2,1x,f10.4,1x,I12)

       open(21, file='input.txt', status='old') ! Open the file with the program parameters.
	   read(21,*) keyword                       ! This reads the filename with the data
       read(21,*) keyword2                      ! Filename with points in 5D space
	   read(21,*) tempword, tempword            ! Reads a line containing headers
       read(21,*) msym, radian, dist_type       ! Specifies the symmetry, units for angles, plot type
       read(21,*) tempword                      ! Read a header line
	   read(21,*) rho_m, rho_p, BallVolumeIndex ! Read the dimensions of the aperture
       read(21,*) name                          ! added to the output filename
       read(21,*) errors                        ! determines whether or not error estimates are written
       read(21,*) numDistinctGBs                ! number of distinct boundaries
       close(21)                                ! Close the input file

       point = index(keyword,'.') - 1

       !this updates the user about the parameters on the consol.
       write(6,5)'I am going to look for data in the file labeled: ', keyword
       write(6,14)'msym =',msym,'radian =',radian
       if (dist_type.eq.1) then
        write(6,6)'the input is being interpreted as grain boundary plane data'
       else
        write(6,6)'the input is being interpreted as grain boundary energy data'
       endif
      write(6,5)'Finding the values at the points listed in: ', keyword2

       ! This part determines the total number of lines
       ! in the data file and assigns it to the variable, nnline
		nnline = 0 ! Initialize counter
		open(41, file=keyword, status='old') ! open the file
 60     continue
		read(41,*,end=61) ! read line, if end of file, goto 61
		nnline=nnline+1  ! Increment counter
		goto 60
 61     continue
        close(41)  ! we are finished, close the file, report result.
		write(6,15) 'There are  ',nnline,' lines in the data file.'
		
        ! Check each line for the string '#'. If found, we know this line
        ! is part of the header.  If not, we have reached the data section.
	    open(41, file=keyword, status='old') ! open the file
        header = 0 ! Initialize counter
        hash = 0   ! Initialize flag
	    do i=1,nnline  ! loop over each line in the file
	     read(41, "(a)") inline    ! reads line into a string
	     hash = index(inline, '#') ! if the character "#" in the string, hash>1.
	     if (hash.ne.0) then       ! check the flag
	      header=header+1          ! when "#" found, it is a header line
		  hash=0                   ! reset the flag
	     else
		  goto 80                  ! when hash is = 0, we have reached the end of the header
	     endif
	    enddo  ! closes i=1,nnline
 80    close(41)				   ! close the file, report the results
	   write(6,15)'There are ',header,' lines in the header'
       write(6,15) 'There are  ',nnline-header,' triangles in the data file.'

       !read in the sampling points
       open(42, file=keyword2, status='old')
       read(42,*)num_misors
       do i=1,num_misors
        read(42,*)ax(i,1),ax(i,2),ax(i,3),ang(i),normals(i,1),normals(i,2),normals(i,3)
       enddo
       close(42)

       !the following are normalization constants for grain boundary plane distribution data
       BallVolume(1) = 0.0000641361 !3° mis, 7° plane
       BallVolume(2) = 0.000139158  !5° mis, 5° plane
       BallVolume(3) = 0.000287439  !5° mis, 7° plane
       BallVolume(4) = 0.00038019   !5° mis, 8° plane
       BallVolume(5) = 0.000484151  !6° mis, 7° plane
       BallVolume(6) = 0.000747069  !7° mis, 7° plane
       BallVolume(7) = 0.00145491   !8° mis, 8° plane

       !This opens the output file, where results will be written
       open(30, file=keyword(1:point)//name//'.txt',status='unknown')
       write(30,5)'# written by kda_pts: ',version
       write(30,5)'# based on data in: ',keyword
       write(30,5)'# using list of points labeled: ',keyword2
       if (dist_type.eq.1) then
        write(30,6)'# this is grain boundary relative area'
       endif
       if (dist_type.eq.2) then
        write(30,6)'# this is grain boundary relative energy'
       endif
       if (dist_type.eq.3) then
        write(30,6)'# this is grain boundary curvature'
       endif
       write(30,"(A,I4)")  '# number of points: ',num_misors
       write(30,6)' angle    value  number'
       !if data on errors was requested, a file is opened for this distribution
       if (errors.eq.1) then
        open(31, file=keyword(1:point)//name//'_errors.txt',status='unknown')
        write(31,"(I4)")  num_misors
       else
        continue
       endif

       call get_symop (msym, O, nsymm)
       !call get_symop (msym, O, nsymm) calls the subroutine that contains the symmetry operators.
       !The variable msym, read from the input file, specifies the symmetry (1 = tetragonal,
       !2 = hexagonal, 3 = cubic, 4 = trigonal, 5 = orthorhombic), O is a matrix containing the
       !operators, and nsymm is the number of symmetry operators.  The routine returns O and nsymm

       !here we have to adjust the normalization factors to account for symmetry
       do i=1,7
        BallVolume(i)=BallVolume(i)*(float(nsymm)/24.0)
       enddo

       rho_m=rho_m*(pi/180.0)  !We have to convert the aperture values to radians
       rho_p=rho_p*(pi/180.0)
       rho_pSq=rho_p*rho_p     !compute rho_p squared - a constant used later

       do i=1,num_misors
        dist(i)=0.0
        err(i)=0.0
        num(i)=0.0
       enddo

       do k=1,num_misors  !this is the main loop over the points in the list
        write(6,*)'work on new point'
        !we start by normalizing the axis and normal vectors.
        len=sqrt(ax(k,1)*ax(k,1)+ax(k,2)*ax(k,2)+ax(k,3)*ax(k,3))
        ax(k,1)=ax(k,1)/len
        ax(k,2)=ax(k,2)/len
        ax(k,3)=ax(k,3)/len
        ang(k)=ang(k)*(pi/180.0)           !convert the misorientation angle to radians
        axis(1)=ax(k,1)
        axis(2)=ax(k,2)
        axis(3)=ax(k,3)
        len=sqrt(normals(k,1)*normals(k,1)+normals(k,2)*normals(k,2)+normals(k,3)*normals(k,3))
        fixedNormal1(1)=normals(k,1)/len
        fixedNormal1(2)=normals(k,2)/len
        fixedNormal1(3)=normals(k,3)/len

        !write(6,25) 'misorientation = ',axis(1), axis(2), axis(3), ang(k)*(180.0/pi)
        !write(6,27) 'normal = ',fixedNormal1(1),fixedNormal1(2),fixedNormal1(3)
        call AAToG (axis, ang(k), gFixed) !convert the axis and angle to an orientation matrix, g.
        call trans (gFixed, gFixedT) !get the transpose

        !In the next section of code, we read the triangles in one at a time.
        !If they are within the aperture, we save their information.
        open(41, file=keyword, status='old') ! open the file and read the header
        do i=1,header
         read(41, "(a)") inline ! read a line of the header
        enddo! ends i=1,header loop

        totalArea = 0
        ct = 0 !introduce a counter
        do i=header+1,nnline !start on the line after the header, continue to end.
         if (dist_type.eq.1) then !This case is for a GB plane distribution
          read(41,*) e1(1), e1(2), e1(3), e2(1), e2(2), e2(3), normal_lab(1), normal_lab(2), normal_lab(3), area
         endif

         if (dist_type.eq.2) then !This case is for a GB energy distribution
          read(41,*) e1(1), e1(2), e1(3), e2(1), e2(2), e2(3), polar, azimuth, dummy, area
          polar=polar*(pi/180.0)
          azimuth=azimuth*(pi/180.0)
          call AnglesToV (polar, azimuth, normal_lab)
         endif

         if (dist_type.eq.3) then !This case is for a GB curvature distribution
          read(41,*) e1(1), e1(2), e1(3), e2(1), e2(2), e2(3), normal_lab(1), normal_lab(2), normal_lab(3), dummy, area
         endif

        if (dist_type.eq.4) then !This case is for a GB velocity distribution
         read(41,*) e1(1), e1(2), e1(3), e2(1), e2(2), e2(3), normal_lab(1), normal_lab(2), normal_lab(3), area_real, curv, vel
        area=abs(vel)
        endif

         ! outputs from some programs produce zero Euler angles when
         ! when one or both orientiations are not known.  These lines are
         ! ignored because they will cause problems later.
         if (e1(1).eq.0.00.and.e1(2).eq.0.00.and.e1(3).eq.0.00) then
          goto 1000 ! Sends it to the next line
         else
          continue
         endif
         if (e2(1).eq.0.00.and.e2(2).eq.0.00.and.e2(3).eq.0.00) then
          goto 1000 ! Sends it to the next line
         else
          continue
         endif

         !if (isnan(normal_lab(1))) then
         !goto 1000
         !endif


         totalArea=totalArea+area !if a good triangle, add the area

         !If the Euler angles are provided in units of degrees, these
         !lines convert them to radians
         if (radian.eq.0) then
          e1(1)=e1(1)*(pi/180.0)
          e1(2)=e1(2)*(pi/180.0)
          e1(3)=e1(3)*(pi/180.0)
          e2(1)=e2(1)*(pi/180.0)
          e2(2)=e2(2)*(pi/180.0)
          e2(3)=e2(3)*(pi/180.0)
         else
         continue
        endif

         ! The Euler angles are convert to 3x3 orientation matrices, g_o1 and g_o2.
         call EToG(e1,g_o1)
         call EToG(e2,g_o2)

         ! Next, we compute delta_m and for all symm equivalents of this
         ! boundary and compare it to the aperture, rho_m.  If it is smaller than
         ! the aperture, we find the normal vectors and record them, with the
         ! area, in the arrary fltd.  This is repeated for every triangle until
         ! the end of the file.

		  do i_sy=1,nsymm             ! loop over all symmetry elements
           call symop(O,i_sy,so1)     ! get first symmetry element for inner loop (so1)
           call MToM (so1,g_o1,gg1)	  ! hit first orientation with a symmetry element
           call MToV(gg1, normal_lab, normal_grain1)

           do j_sy=1,nsymm             ! loop over all symmetry elements
		    call symop(O,j_sy,so2)     ! get the second symmetry element for outer loop (so2)
		    call MToM(so2,g_o2,gg2)	   ! hit second orientation with a symmetry element
		    call trans(gg2,g2sT)       ! take the transpose of the second orientation
		    call MToM(gg1,g2sT,dg)     ! calculate the misorientation (dg)
            call trans(dg,dgT)
            call MToM(dg,gFixedT,diffFromFixed)
            tem=(diffFromFixed(1,1)+diffFromFixed(2,2)+diffFromFixed(3,3)-1)/2
            if (tem.gt.1.0) tem=1.0
            if (tem.lt.-1.0) tem=-1.0
		    delta_m=acos(tem)
            if (delta_m.lt.rho_m) then  ! check to see if it is within the aperture
             ct=ct+1 !increment ct everytime a triangle is within the aperture
             call MToV(dgT, normal_grain1, normal_grain2)
             ftrd(ct,1)=normal_grain1(1)
             ftrd(ct,2)=normal_grain1(2)
             ftrd(ct,3)=normal_grain1(3)
             ftrd(ct,4)=-normal_grain2(1)
             ftrd(ct,5)=-normal_grain2(2)
             ftrd(ct,6)=-normal_grain2(3)
             ftrd(ct,7)=area
            else
             continue
            endif
            call MToM(dgT,gFixedT,diffFromFixed)
            tem=(diffFromFixed(1,1)+diffFromFixed(2,2)+diffFromFixed(3,3)-1)/2
            if (tem.gt.1.0) tem=1.0
            if (tem.lt.-1.0) tem=-1.0
            delta_m=acos(tem)
            if (delta_m.lt.rho_m) then  ! check to see if it is within the aperture
             ct=ct+1 !increment ct everytime a triangle is within the aperture
             call MToV(dgT, normal_grain1, normal_grain2)
             ftrd(ct,1)=-normal_grain2(1)
             ftrd(ct,2)=-normal_grain2(2)
             ftrd(ct,3)=-normal_grain2(3)
             ftrd(ct,4)=normal_grain1(1)
             ftrd(ct,5)=normal_grain1(2)
             ftrd(ct,6)=normal_grain1(3)
             ftrd(ct,7)=area
            else
             continue
            endif
		   enddo ! closes j_sy=1,nsymm
		  enddo ! closes i_sy=1,nsymm
          !this is a progress update
		  if (mod(i,250000).eq.0) then
	       write(6,16) 'classified ',i, ' of ',nnline,' triangles'
	      endif


 1000     continue ! so we go to the next line of the data file and repeat
         enddo    ! This closes i=header+1,nnline, the loop over each line of data.

         write(6,15)'there are ',ct,' triangles within the misorientation aperture'
         close(41)


          call MToV(gFixedT, fixedNormal1, fixedNormal2)
          sum = 0.0
          numTris = 0
          do j=1,ct
           normal_grain1(1)=ftrd(j,1)
           normal_grain1(2)=ftrd(j,2)
           normal_grain1(3)=ftrd(j,3)
           normal_grain2(1)=ftrd(j,4)
           normal_grain2(2)=ftrd(j,5)
           normal_grain2(3)=ftrd(j,6)
           area=ftrd(j,7)
           sign=1.0
           chi1=acos(sign*( (normal_grain1(1)*fixedNormal1(1)) + (normal_grain1(2)*fixedNormal1(2)) + (normal_grain1(3)*fixedNormal1(3)) ))
           chi2=acos(-sign*( (normal_grain2(1)*fixedNormal2(1)) + (normal_grain2(2)*fixedNormal2(2)) + (normal_grain2(3)*fixedNormal2(3)) ))
           delta_pSq=0.5*((chi1*chi1)+(chi2*chi2))
           if (delta_pSq.lt.rho_pSq) then
            numTris=numTris+1
            sum = sum + area
           else
            continue
           endif
           sign=-1.0
           chi1=acos(sign*( (normal_grain1(1)*fixedNormal1(1)) + (normal_grain1(2)*fixedNormal1(2)) + (normal_grain1(3)*fixedNormal1(3)) ))
           chi2=acos(-sign*( (normal_grain2(1)*fixedNormal2(1)) + (normal_grain2(2)*fixedNormal2(2)) + (normal_grain2(3)*fixedNormal2(3)) ))
           delta_pSq=0.5*((chi1*chi1)+(chi2*chi2))
           if (delta_pSq.lt.rho_pSq) then
            numTris=numTris+1
            sum = sum + area
           else
            continue
           endif
          enddo !j loop over triangles

          dist(k)=sum ! assigns the total area at this sample point
          num(k)=numTris ! assigns number of energy values added at this sample point
          write(6,7) 'finished point ',k,' of ',num_misors

        if (dist_type.eq.1) then
         dist(k)=dist(k)/totalArea
         dist(k) = dist(k)/BallVolume(BallVolumeIndex)
        else
         dist(k)=dist(k)/num(k)
        endif
        write(6,28)ang(k)*(180.0/pi),dist(k),int(num(k))

        enddo !this is the k loop

         !write result at each sampling point
         do i=1,num_misors
          write (30,28)ang(i)*(180.0/pi),dist(i),int(num(i))
          enddo
         close(30) ! closes the output file

         if (errors.eq.1) then  ! Write a file with errors, when requested
          do i=1, num_samplPts
           err(i)=sqrt((dist(i)/totalArea/numDistinctGBs))/BallVolume(BallVolumeIndex)
          enddo
          do i=1,num_misors
           write (31,*)ang(i),err(i)
          enddo
          close(31) ! closes the output file
         else
          continue
         endif

 9000    continue

         end






	   
	   
	   
	   
	   
	   
	   
	   
	   
	   
	   
	   
	   
	   
	   
	   
	   
	   
	   
	   
	   