C*****************************************
       program disorientation calculator
C*****************************************
       implicit none
	   include 'common.fi'

       write(6,*) '================================================='
	   write(6,*) ' PROGRAM USED TO TO DETERMINE DISORIENTATIONS '
	   write(6,*) ' AND DISORIENATION DISTRIBUTIONS IN ORTHORHOMBIC,'
	   write(6,*) ' TRIGONAL, TETRAGONAL, HEXAGONAL, AND CUBIC '
	   write(6,*) ' SYSTEMS '
	   write(6,*) ' '
	   write(6,*) ' rohrer@cmu.edu'
	   write(6,*) ' version 01/22/2019'
       write(6,*) '================================================='
	   
	   version = 'version 01/22/2019'

       !some constants that might be useful
       pi = 4.0*atan(1.0)
       !some formating statements used to format typed output
  5    format(A,A)
  6    format(A)
  13   format(A,f4.1)
  14   format(A,I2,2X,A,I2)
  15   format(A,I8,A)
  17   format(f5.1,6x,F6.4,5x,F6.4)
  18   format(A,f6.3,A,f6.3,A,f6.3,A,f6.3,A,f6.3,A,3f6.3)
  19   format(f5.2,A,f5.2,A,f5.2,A,f5.2,A)
  20   format(A,f6.3,A,f6.3,A,f6.3)
  21   format(A,f6.3,2f7.3,A)

       open(21, file='input.txt', status='old') ! Open the file with the program
                                                ! parameters.
	   read(21,*) keyword                       ! This reads the filename with the data
       read(21,*) mode                          ! Specifies the type of input (single, 2D, 3D)
	   read(21,*) tempword, tempword            ! Reads a line containing headers
       read(21,*) msym, radian                  ! Specifies the symmetry and units for angles
       read(21,*) tempword                      ! Read a header line
	   read(21,*) bin                           ! Read the number of degrees per bin

       call get_symop (msym, O, nsymm)
       !call get_symop (msym, O, nsymm) calls the subroutine that contains the symmetry operators.
       !The variable msym, read from the input file, specifies the symmetry (1 = tetragonal,
       !2 = hexagonal, 3 = cubic, 4 = trigonal, 5 = orthorhombic), O is a matrix containing the
       !operators, and nsymm is the number of symmetry operators.  The routine returns O and nsymm

       if (mode.gt.2) goto 2000  ! if single disorientation calculation rather than a
                                 ! distribution, go to that part of the code.
       ! This defines the maximum angle for a disorientation distribution
       if (msym.eq.1) limit=99  !max eq 98.42,  D4, 422, 4/m mm
	   if (msym.eq.2) limit=94  !max eq 93.84,  D6, 622, 6/m mm	   
	   if (msym.eq.3) limit=63  !max eq 62.7994, O, 432, m-3m
	   if (msym.eq.4) limit=105 !max eq 104.5,  D3,  32, -3m
	   if (msym.eq.5) limit=120 !max eq 120.0,  D2, 222, mmm
	   
	   bin_num = int(limit/bin) !computes the number of bins in the distribution
       !repeats back to you the parameters you specified in the input.
       write(6,5)'I am going to look for data in the file labeled: ', keyword
       write(6,14)'msym =',msym,'radian =',radian
       write(6,13)'degrees per bin =',bin
       close(21)

       do i=1,2000  ! This just ensures the initial values of the distribution are zero
        d(i,1)=0.0
        d(i,2)=0.0
       enddo
		
       ! This part is to determines the total number of lines
       ! in the data file and assigns it to the variable, nnline
		nnline = 0 ! Initialize counter
		open(41, file=keyword, status='old') ! open the file
 60     continue
		read(41,*,end=61) ! read line, if end of file, goto 61
		nnline=nnline+1  ! Increment counter
		goto 60
 61     continue
        close(41)  ! we are finished, close the file
		write(6,15) 'There are  ',nnline,' lines in the data file.'
		
        ! Check each line for the string '#'. If found, we know this line
        ! is part of the header.  If not, we have reached the data section.
	    open(41, file=keyword, status='old') ! open the file
        header = 0 ! Initialize counter
        hash = 0   ! Initialize flag
	    do i=1,nnline  ! loop over each line in the file
	     read(41, "(a)") inline    ! reads line into a string
	     hash = index(inline, '#') ! if the character "#" in the string, hash>1.
	     if (hash.ne.0) then       ! check the flag
	      header=header+1          ! when "#" found, it is a header line
		  hash=0                   ! reset the flag
	     else
		  goto 80                  ! when hash is = 0, we have reached the end of the header
	     endif
	    enddo  ! closes i=1,nnline
 80    close(41)				   ! close the file
	   write(6,*)header,' lines in the header'	
	   ! This opens the output file, where results are written
       open(30, file='report_'//keyword, status='unknown')
       ! Now we read the data
       open(41, file=keyword, status='old') ! Open the input file
       nline = 0 ! Before reading the data, we have to read the neader
       do i=1,header
        read(41, "(a)") inline ! read a line of the header
        write(30, "(a)")inline ! write that line into the output file
       enddo! closes i=1,header
       blank = 0  ! Initialize counter for zero aEuler angle lines

       ! Next is the start of the main loop, reading reading and
       ! analyzing one line of data at a time.
       do kk=header+1,nnline !start on the line after the header, continue to end.

        ! The same code is used to compute the disorientation distribution from
        ! 2D and 3D data.  The only differences are the structure of the input
        ! file and boundary area (3D) versus length (2D). Here is where we decide
        ! which type of file to read.
        if (mode.eq.1) then
         ! Expect first three values are first grain Euler angles (e1),
         ! second three values are second grain Euler angles (e2),
         ! then the x, y, and z components of the normal vector, then the triangle area
		 read(41,*) e1(1), e1(2), e1(3), e2(1), e2(2), e2(3),
     &	 normal(1), normal(2), normal(3), area
        else
         ! Expect first three values are first grain Euler angles (e1),
         ! second three values are second grain Euler angles (e2), two
         ! parameters we ignore, then the x,y coorinates of the grain 
         ! boundary trace end points.
         read(41,*) e1(1), e1(2), e1(3), e2(1), e2(2), e2(3),
     &	 dummy, dummy, x1, y1, x2, y2
         nl(1)=x2-x1 !these three lines compute the length
		 nl(2)=y2-y1
	     len=sqrt((nl(1)*nl(1))+(nl(2)*nl(2)))
         area=len !for now, call it area
        endif

        ! Outputs from some programs produce zero Euler angles when the
        ! when one or both orientiations are not known.  These lines are
        ! ignored because they will cause problems later.
		if (e1(1).eq.0.00.and.e1(2).eq.0.00.and.e1(3).eq.0.00) then
          blank=blank+1 ! increment counter
		  goto 1000 ! Sends it to the next line
        else
          continue
		endif
        if (e2(1).eq.0.00.and.e2(2).eq.0.00.and.e2(3).eq.0.00) then
          blank=blank+1 ! increment counter
          goto 1000 ! Sends it to the next line
        else
          continue
        endif
        ! If the Euler angles are provided in units of degrees, these
		! lines convert them to radians
	    if (radian.eq.0) then
	     e1(1)=e1(1)*(pi/180.0)
	     e1(2)=e1(2)*(pi/180.0)
	     e1(3)=e1(3)*(pi/180.0)
	     e2(1)=e2(1)*(pi/180.0)
	     e2(2)=e2(2)*(pi/180.0)
	     e2(3)=e2(3)*(pi/180.0)
        else
         continue
	    endif

        ! The Euler angles are convert to 3x3 orientation matrices, g_o1 and g_o2.
        call EToG(e1,g_o1)
		call EToG(e2,g_o2)

        ! These nested loops check all indistinguishable representations and 
        ! identify the minimum misorientation (disorientation) for each pair 
        ! of orientations
		disor2 = 3.1415 ! Start the parameter at the max disorientation
		do i_sy=1,nsymm               ! loop over all symetry elements
		  do j_sy=1,nsymm             ! loop over all symetry elements
		   call symop(O,i_sy,so1)     ! get first symmetry element for inner loop (so1)
		   call MToM (so1,g_o1,gg1)	  ! hit first orientation with a symmetry element
		   call symop(O,j_sy,so2)     ! get the second symmetry element for outer loop (so2)
		   call MToM(so2,g_o2,gg2)	  ! hit second orientation with a symmetry element
		   call trans(gg2,gtr)        ! take the transpose of the second orientation
		   call MToM(gg1,gtr,dg)      ! calculate the misorientation (dg)
		   misor=acos((dg(1,1)+dg(2,2)+dg(3,3)-1)/2)  ! calculate the misorietation angle
           if (misor.lt.disor2) then  ! if it is less than the smallest angle found so far
            disor2=misor              ! then remember that value in disor2
           endif
		   call trans (gg1, gtr)      ! reverse orientations and repeat; transpose first orientation
		   call MToM (gg2, gtr, dg)   ! calculate the misorientation with orientations reversed
		   misor=acos((dg(1,1)+dg(2,2)+dg(3,3)-1)/2)  ! calculate the misorietation angle
		   if (misor.lt.disor2) then  ! if it is less than the smallest angle found so far
			disor2=misor              ! then remember that value in disor2
		   endif
		  enddo ! closes j_sy=1,nsymm
		 enddo ! closes i_sy=1,nsymm
         ! When we exit this loop, the parameter disor2 holds the smallest angle found.
		 
         ! The next section bins the observation in the matix d.  
         ! The first column of d accumulates the areas of the 
         ! trianges and the second the number accumulates the
         ! numnber of triangles.
		 disor2=disor2*(180.0/pi)     ! convert the value to degrees
		 ind = floor(disor2/bin)+1    ! finds index of d (smallest integer > 0 for bin number)
		 if (ind.eq.0) ind=1          ! make sure the index is non-zero
		 if (ind.gt.bin_num) then     ! make sure it is not > the uper limit
		  ind=bin_num
		  write(6,*)'misorientation out of bounds'
		 endif 
		 d(ind,1)=d(ind,1)+area       ! add the area to the first column of d
		 d(ind,2)=d(ind,2)+1.0        ! increment the second column of d
         area_total=area_total+area   ! keep track ot the total area of all triangles
         ! when you reach this point, the orientation pair has been classified
 1000    continue ! so we go to the next line of the data file and repeat
       enddo    ! This closes kk=header+1,nnline, the loop over each line of data.

         ct=nnline-header-blank         ! ct is the lines of useful data
         ! The next four lines normalize the data
         do i=0,bin_num                 ! loop over each bin
          d_norm(i,1)=d(i,1)/area_total ! divide area by the total area
          d_norm(i,2)=d(i,2)/float(ct)  ! divide number by the total number
         enddo
         ! This section writes the output to the terminal and to a file
         ! These statements write header information
         write(6,6)' '
         write(30,6)'# '
         if (mode.eq.1) then  ! The output is different for 2D and 3D data
          write(6,6)'Disorientation distribution: rotation angle, area fraction, and number fraction'
          write(30,6)'# Disorientation distribution: rotation angle, area fraction, and number fraction'
         else
          write(6,6)'Disorientation distribution: rotation angle, length fraction, and number fraction'
          write(30,6)'# Disorientation distribution: rotation angle, length fraction, and number fraction'
         endif
         write(6,6)' '
         write(30,6)'# '
         if (mode.eq.1) then  ! The output is different for 2D and 3D data
          write(6,6)'degrees    area       number'
          write(30,6)'  degrees  area       number'
         else
          write(6,6)'degrees    length     number'
          write(30,6)'  degrees  length     number'
         endif
         ! These statements write each line of the distribution
         do i=0,bin_num
		  write(30,17)i*bin, d_norm(i,1),d_norm(i,2)
          write(6,17)i*bin, d_norm(i,1),d_norm(i,2)
         enddo ! closes the i=0,bin_num loop
 
         close(30) ! closes the output file
         goto 9000  ! this is the end of mode 1/ mode 2 calculations

 2000    continue

         if (mode.eq.3) then !Here, the mode 3 calculation is made
          open(41, file=keyword, status='old') ! Open the input file
          read(41,*) e1(1), e1(2), e1(3), e2(1), e2(2), e2(3)
          ! If the Euler angles are provided in units of degrees, these
		  ! lines convert them to radians
	      if (radian.eq.0) then
	       e1(1)=e1(1)*(pi/180.0)
	       e1(2)=e1(2)*(pi/180.0)
	       e1(3)=e1(3)*(pi/180.0)
	       e2(1)=e2(1)*(pi/180.0)
	       e2(2)=e2(2)*(pi/180.0)
	       e2(3)=e2(3)*(pi/180.0)
          else
           continue
	      endif
          ! The Euler angles are convert to 3x3 orientation matrices, g_o1 and g_o2.
          call EToG(e1,g_o1)
		  call EToG(e2,g_o2)
          ! These nested loops check all indistinguishable representations and
          ! identify the minimum misorientation (disorientation) for each pair
          ! of orientations
		  disor2 = 3.1415 ! Start the parameter at the max disorientation
		  do i_sy=1,nsymm               ! loop over all symetry elements
           do j_sy=1,nsymm             ! loop over all symetry elements
		    call symop(O,i_sy,so1)     ! get first symmetry element for inner loop (so1)
		    call MToM (so1,g_o1,gg1)	  ! hit first orientation with a symmetry element
		    call symop(O,j_sy,so2)     ! get the second symmetry element for outer loop (so2)
		    call MToM(so2,g_o2,gg2)	  ! hit second orientation with a symmetry element
		    call trans(gg2,gtr)        ! take the transpose of the second orientation
		    call MToM(gg1,gtr,dg)      ! calculate the misorientation (dg)
		    misor=acos((dg(1,1)+dg(2,2)+dg(3,3)-1)/2)  ! calculate the misorietation angle
            if (misor.lt.disor2) then  ! if it is less than the smallest angle found so far
             disor2=misor              ! then remember that value in disor2
             g_d=dg                    ! remember the disorientation matrix
            endif
		    call trans (gg1, gtr)      ! reverse orientations and repeat; transpose first orientation
		    call MToM (gg2, gtr, dg)   ! calculate the misorientation with orientations reversed
		    misor=acos((dg(1,1)+dg(2,2)+dg(3,3)-1)/2)  ! calculate the misorietation angle
            if (misor.lt.disor2) then  ! if it is less than the smallest angle found so far
             disor2=misor              ! then remember that value in disor2
             g_d=dg                    ! remember the disorientation matrix
            endif
		   enddo ! closes j_sy=1,nsymm
		   enddo ! closes i_sy=1,nsymm
           ! When we exit this loop, the parameter disor2 holds the smallest angle found
           ! and g_d holds the misorientation matrix
           call GToE(dg,e3)
           ax(1)=(g_d(2,3)-g_d(3,2))/(2*sin(disor2))     ! calculate the misorietation axis
           ax(2)=(g_d(3,1)-g_d(1,3))/(2*sin(disor2))
           ax(3)=(g_d(1,2)-g_d(2,1))/(2*sin(disor2))
           ! To finish mode 3, we write out the answers
           write(6,18)'The input angles are: ',e1(1),',',e1(2),',',e1(3),' and',e2(1),',',e2(2),',',e2(3)
           write(6,6)'The axis-angle representation for this disorientation is:'
           write(6,19)disor2*(180.0/pi),'° around [',ax(1),',',ax(2),',',ax(3),']'
           write(6,20)'The Euler angles for this disorientation are: ',e3(1),',',e3(2),',',e3(3)
           write(6,6)'The 3x3 disorientation matrix is:'
           write(6,21)'|',g_d(1,1),g_d(1,2),g_d(1,3),'|'
           write(6,21)'|',g_d(2,1),g_d(2,2),g_d(2,3),'|'
           write(6,21)'|',g_d(3,1),g_d(3,2),g_d(3,3),'|'
         else
          goto 2100
         endif
         goto 9000  ! this is the end of mode 3 calculations

 2100    continue
         if (mode.eq.4) then !Here, the mode 4 calculation is made
          open(41, file=keyword, status='old') ! Open the input file
          read(41,*) ax(1), ax(2), ax(3), misor
          ss=sqrt(ax(1)*ax(1)+ax(2)*ax(2)+ax(3)*ax(3))
          ax(1)=ax(1)/ss
          ax(2)=ax(2)/ss
          ax(3)=ax(3)/ss
          misor=misor*(pi/180.0)
          call AAToG (ax, misor, g_d)
          call GToE(g_d,e3)
          ! To finish mode 4, we write out the answers
          write(6,6)'The input misorientation is: '
          write(6,19)misor*(180.0/pi),'° around [',ax(1),',',ax(2),',',ax(3),']'
          write(6,20)'The Euler angles for this disorientation are: ',e3(1),',',e3(2),',',e3(3)
          write(6,6)'The 3x3 disorientation matrix is:'
          write(6,21)'|',g_d(1,1),g_d(1,2),g_d(1,3),'|'
          write(6,21)'|',g_d(2,1),g_d(2,2),g_d(2,3),'|'
          write(6,21)'|',g_d(3,1),g_d(3,2),g_d(3,3),'|'
         else
          goto 2200
         endif
         goto 9000  ! this is the end of mode 4 calculations

 2200    continue
         if (mode.eq.5) then !Here, the mode 5 calculation is made
          open(41, file=keyword, status='old') ! Open the input file
          read(41,*) g_d(1,1),g_d(1,2),g_d(1,3)
          read(41,*) g_d(2,1),g_d(2,2),g_d(2,3)
          read(41,*) g_d(3,1),g_d(3,2),g_d(3,3)
          disor2=acos((g_d(1,1)+g_d(2,2)+g_d(3,3)-1)/2)  ! calculate the misorietation angle
          ax(1)=(g_d(2,3)-g_d(3,2))/(2*sin(disor2))     ! calculate the misorietation axis
          ax(2)=(g_d(3,1)-g_d(1,3))/(2*sin(disor2))
          ax(3)=(g_d(1,2)-g_d(2,1))/(2*sin(disor2))
          call GToE(g_d,e3)
          ! To finish mode 5, we write out the answers
          write(6,6)'The input misorientation is: '
          write(6,21)'|',g_d(1,1),g_d(1,2),g_d(1,3),'|'
          write(6,21)'|',g_d(2,1),g_d(2,2),g_d(2,3),'|'
          write(6,21)'|',g_d(3,1),g_d(3,2),g_d(3,3),'|'
          write(6,6)'The axis-angle representation for this misorientation is:'
          write(6,19)disor2*(180.0/pi),'° around [',ax(1),',',ax(2),',',ax(3),']'
          write(6,20)'The Euler angles for this disorientation are: ',e3(1),',',e3(2),',',e3(3)
         else
          goto 2300
         endif
         goto 9000  ! this is the end of mode 5 calculations

 2300    continue
         if (mode.eq.6) then !Here, the mode 6 calculation is made
          open(41, file=keyword, status='old') ! Open the input file
          read(41,*) e3(1), e3(2), e3(3)
          call EToG(e3,g_d)
          disor2=acos((g_d(1,1)+g_d(2,2)+g_d(3,3)-1)/2)  ! calculate the misorietation angle
          ax(1)=(g_d(2,3)-g_d(3,2))/(2*sin(disor2))     ! calculate the misorietation axis
          ax(2)=(g_d(3,1)-g_d(1,3))/(2*sin(disor2))
          ax(3)=(g_d(1,2)-g_d(2,1))/(2*sin(disor2))
          ! To finish mode 6, we write out the answers
          write(6,20)'The input Euler angles for this misorientation are: ',e3(1),',',e3(2),',',e3(3)
          write(6,6)'The axis-angle representation for this disorientation is:'
          write(6,19)disor2*(180.0/pi),'° around [',ax(1),',',ax(2),',',ax(3),']'
          write(6,6)'The 3x3 disorientation matrix is:'
          write(6,21)'|',g_d(1,1),g_d(1,2),g_d(1,3),'|'
          write(6,21)'|',g_d(2,1),g_d(2,2),g_d(2,3),'|'
          write(6,21)'|',g_d(3,1),g_d(3,2),g_d(3,3),'|'
         else
          goto 9000
         endif
         goto 9000  ! this is the end of mode 6 calculations


 9000    continue

       end






	   
	   
	   
	   
	   
	   
	   
	   
	   
	   
	   
	   
	   
	   
	   
	   
	   
	   
	   
	   
	   